'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var react = require('react');

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _createForOfIteratorHelperLoose(o, allowArrayLike) {
  var it;

  if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) {
    if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") {
      if (it) o = it;
      var i = 0;
      return function () {
        if (i >= o.length) return {
          done: true
        };
        return {
          done: false,
          value: o[i++]
        };
      };
    }

    throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
  }

  it = o[Symbol.iterator]();
  return it.next.bind(it);
}

var hasInitialValue = function hasInitialValue(atom) {
  return 'init' in atom;
};

var concatMap = function concatMap(m1, m2) {
  // Map([...m1, ...m2]) alternative
  var newMap = new Map(m1);
  m2.forEach(function (value, key) {
    newMap.set(key, value);
  });
  return newMap;
};

var createState = function createState(initialValues, newAtomReceiver) {
  var currWip = new Map();
  var queue = [];

  var updateState = function updateState(updater) {
    queue.push(updater);

    if (queue.length > 1) {
      return;
    }

    var nextWip = currWip;

    while (queue.length) {
      nextWip = queue[0](nextWip);
      queue.shift();
    }

    if (nextWip !== currWip) {
      currWip = nextWip;

      if (currWip.size) {
        var atomsToNotify = new Set(currWip.keys());
        mountDependencies(state, currWip);
        commitState(state, currWip);
        ++state.v;
        atomsToNotify.forEach(function (atom) {
          var mounted = state.m.get(atom);
          mounted == null ? void 0 : mounted.l.forEach(function (listener) {
            return listener();
          });
        });
      }
    }
  };

  var state = {
    n: newAtomReceiver,
    v: 0,
    a: new WeakMap(),
    m: new WeakMap(),
    u: updateState
  };

  if (initialValues) {
    for (var _iterator = _createForOfIteratorHelperLoose(initialValues), _step; !(_step = _iterator()).done;) {
      var _step$value = _step.value,
          atom = _step$value[0],
          value = _step$value[1];
      var atomState = {
        v: value,
        r: 0,
        d: new Map()
      };

      if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
        Object.freeze(atomState);
      }

      state.a.set(atom, atomState);
    }
  }

  return state;
};

var getAtomState = function getAtomState(state, wip, atom) {
  return wip.get(atom) || state.a.get(atom);
};

var wipAtomState = function wipAtomState(state, wip, atom, dependencies, promise) {
  var atomState = getAtomState(state, wip, atom);

  if (promise && promise !== (atomState == null ? void 0 : atomState.p)) {
    // newer async read is running, not updating
    return null;
  }

  var nextAtomState = _extends({
    r: 0
  }, atomState, {
    d: dependencies ? new Map(Array.from(dependencies).map(function (a) {
      var _getAtomState$r, _getAtomState;

      return [a, (_getAtomState$r = (_getAtomState = getAtomState(state, wip, a)) == null ? void 0 : _getAtomState.r) != null ? _getAtomState$r : 0];
    })) : atomState ? atomState.d : new Map()
  });

  if (!atomState && hasInitialValue(atom)) {
    nextAtomState.v = atom.init;
  }

  return nextAtomState;
};

var setAtomValue = function setAtomValue(state, wip, atom, value, dependencies, promise) {
  var atomState = wipAtomState(state, wip, atom, dependencies, promise);

  if (!atomState) {
    return wip;
  }

  delete atomState.e;
  delete atomState.p;

  if (!('v' in atomState) || !Object.is(atomState.v, value)) {
    atomState.v = value;
    ++atomState.r;
  }

  return new Map(wip).set(atom, atomState);
};

var setAtomReadError = function setAtomReadError(state, wip, atom, error, dependencies, promise) {
  var atomState = wipAtomState(state, wip, atom, dependencies, promise);

  if (!atomState) {
    return wip;
  }

  delete atomState.p;
  atomState.e = error;
  return new Map(wip).set(atom, atomState);
};

var setAtomReadPromise = function setAtomReadPromise(state, wip, atom, promise, dependencies) {
  var atomState = wipAtomState(state, wip, atom, dependencies);
  atomState.p = promise;
  return new Map(wip).set(atom, atomState);
};

var setAtomWritePromise = function setAtomWritePromise(state, wip, atom, promise) {
  var atomState = wipAtomState(state, wip, atom);

  if (promise) {
    atomState.w = promise;
  } else {
    delete atomState.w;
  }

  return new Map(wip).set(atom, atomState);
};

var scheduleReadAtomState = function scheduleReadAtomState(state, atom, promise) {
  promise.then(function () {
    state.u(function (wip) {
      return readAtomState(state, wip, atom, true)[1];
    });
  });
};

var readAtomState = function readAtomState(state, wip, atom, force) {
  if (!force) {
    var atomState = getAtomState(state, wip, atom);

    if (atomState && Array.from(atomState.d.entries()).every(function (_ref) {
      var a = _ref[0],
          r = _ref[1];
      var aState = getAtomState(state, wip, a);
      return aState && !aState.e && !aState.p && aState.r === r;
    })) {
      return [atomState, wip];
    }
  }

  var error;
  var promise;
  var value;
  var dependencies = new Set();

  try {
    var promiseOrValue = atom.read(function (a) {
      dependencies.add(a);

      if (a !== atom) {
        var _aState;

        var _readAtomState = readAtomState(state, wip, a);

        _aState = _readAtomState[0];
        wip = _readAtomState[1];

        if (_aState.e) {
          throw _aState.e; // read error
        }

        if (_aState.p) {
          throw _aState.p; // read promise
        }

        return _aState.v; // value
      } // a === atom


      var aState = getAtomState(state, wip, a);

      if (aState) {
        if (aState.p) {
          throw aState.p; // read promise
        }

        return aState.v; // value
      }

      if (hasInitialValue(a)) {
        return a.init;
      }

      throw new Error('no atom init');
    });

    if (promiseOrValue instanceof Promise) {
      promise = promiseOrValue.then(function (value) {
        state.u(function (prev) {
          return setAtomValue(state, concatMap(prev, wip), atom, value, dependencies, promise);
        });
      }).catch(function (e) {
        if (e instanceof Promise) {
          scheduleReadAtomState(state, atom, e);
          return e;
        }

        state.u(function (prev) {
          return setAtomReadError(state, concatMap(prev, wip), atom, e instanceof Error ? e : new Error(e), dependencies, promise);
        });
      });
    } else {
      value = promiseOrValue;
    }
  } catch (errorOrPromise) {
    if (errorOrPromise instanceof Promise) {
      scheduleReadAtomState(state, atom, errorOrPromise);
      promise = errorOrPromise;
    } else if (errorOrPromise instanceof Error) {
      error = errorOrPromise;
    } else {
      error = new Error(errorOrPromise);
    }
  }

  if (error) {
    wip = setAtomReadError(state, wip, atom, error, dependencies);
  } else if (promise) {
    wip = setAtomReadPromise(state, wip, atom, promise, dependencies);
  } else {
    wip = setAtomValue(state, wip, atom, value, dependencies);
  }

  var ret = [getAtomState(state, wip, atom), wip];
  wip = new Map(); // for async

  return ret;
};

var readAtom = function readAtom(state, readingAtom) {
  var _readAtomState2 = readAtomState(state, new Map(), readingAtom),
      atomState = _readAtomState2[0],
      wip = _readAtomState2[1]; // schedule commit


  if (wip.size) {
    state.u(function (prev) {
      // XXX this is very tricky, any idea to improve?
      mountDependencies(state, wip);
      commitState(state, wip);
      return prev;
    });
  }

  return atomState;
};

var addAtom = function addAtom(state, addingAtom) {
  var mounted = state.m.get(addingAtom);

  if (!mounted) {
    mounted = mountAtom(state, new Map(), addingAtom);
  }

  return mounted;
}; // XXX doesn't work with mutally dependent atoms


var canUnmountAtom = function canUnmountAtom(atom, mounted) {
  return !mounted.l.size && (!mounted.d.size || mounted.d.size === 1 && mounted.d.has(atom));
};

var delAtom = function delAtom(state, deletingAtom) {
  var mounted = state.m.get(deletingAtom);

  if (mounted && canUnmountAtom(deletingAtom, mounted)) {
    unmountAtom(state, new Map(), deletingAtom);
  }
};

var updateDependentsState = function updateDependentsState(state, wip, atom, prevAtomState) {
  var _getAtomState2;

  if (prevAtomState && !prevAtomState.e && !prevAtomState.p && prevAtomState.r === ((_getAtomState2 = getAtomState(state, wip, atom)) == null ? void 0 : _getAtomState2.r)) {
    return wip; // bail out
  }

  var mounted = state.m.get(atom);
  mounted == null ? void 0 : mounted.d.forEach(function (dependent) {
    var _nextDependentState$p;

    if (dependent === atom) {
      return;
    }

    var dependentState = getAtomState(state, wip, dependent);
    var nextDependentState;

    var _readAtomState3 = readAtomState(state, wip, dependent, true);

    nextDependentState = _readAtomState3[0];
    wip = _readAtomState3[1];
    (_nextDependentState$p = nextDependentState.p) == null ? void 0 : _nextDependentState$p.then(function () {
      state.u(function (prev) {
        return updateDependentsState(state, prev, dependent);
      });
    });
    wip = updateDependentsState(state, wip, dependent, dependentState);
  });
  return wip;
};

var writeAtomState = function writeAtomState(state, wip, atom, update, pendingPromises) {
  var atomState = getAtomState(state, wip, atom);

  if (atomState && atomState.w) {
    var promise = atomState.w.then(function () {
      state.u(function (prev) {
        return writeAtomState(state, prev, atom, update);
      });
    });

    if (pendingPromises) {
      pendingPromises.push(promise);
    }

    return wip;
  }

  var isSync = true;

  try {
    var promiseOrVoid = atom.write(function (a) {
      var _readAtomState4 = readAtomState(state, wip, a),
          aState = _readAtomState4[0];

      if (aState.e) {
        throw aState.e; // read error
      }

      if (aState.p) {
        if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
          console.warn('Reading pending atom state in write operation. We throw a promise for now.', a);
        }

        throw aState.p; // read promise
      }

      if ('v' in aState) {
        return aState.v;
      }

      if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
        console.warn('[Bug] no value found while reading atom in write operation. This probably a bug.', a);
      }

      throw new Error('no value found');
    }, function (a, v) {
      if (a === atom) {
        var aState = getAtomState(state, wip, a);

        if (isSync) {
          wip = updateDependentsState(state, setAtomValue(state, wip, a, v), a, aState);
        } else {
          state.u(function (prev) {
            return updateDependentsState(state, setAtomValue(state, prev, a, v), a, aState);
          });
        }
      } else {
        if (isSync) {
          wip = writeAtomState(state, wip, a, v);
        } else {
          state.u(function (prev) {
            return writeAtomState(state, prev, a, v);
          });
        }
      }
    }, update);

    if (promiseOrVoid instanceof Promise) {
      if (pendingPromises) {
        pendingPromises.push(promiseOrVoid);
      }

      wip = setAtomWritePromise(state, wip, atom, promiseOrVoid.then(function () {
        state.u(function (prev) {
          return setAtomWritePromise(state, prev, atom);
        });
      }));
    }
  } catch (e) {
    if (pendingPromises && pendingPromises.length) {
      pendingPromises.push(new Promise(function (_resolve, reject) {
        reject(e);
      }));
    } else {
      throw e;
    }
  }

  isSync = false;
  return wip;
};

var writeAtom = function writeAtom(state, writingAtom, update) {
  var pendingPromises = [];
  state.u(function (prev) {
    return writeAtomState(state, prev, writingAtom, update, pendingPromises);
  });

  if (pendingPromises.length) {
    return new Promise(function (resolve, reject) {
      var loop = function loop() {
        var len = pendingPromises.length;

        if (len === 0) {
          resolve();
        } else {
          Promise.all(pendingPromises).then(function () {
            pendingPromises.splice(0, len);
            loop();
          }).catch(reject);
        }
      };

      loop();
    });
  }
};

var isActuallyWritableAtom = function isActuallyWritableAtom(atom) {
  return !!atom.write;
};

var mountAtom = function mountAtom(state, wip, atom, initialDependent) {
  // mount dependencies beforehand
  var atomState = getAtomState(state, wip, atom);

  if (atomState) {
    atomState.d.forEach(function (_, a) {
      if (a !== atom) {
        // check if not mounted
        if (!state.m.has(a)) {
          mountAtom(state, wip, a, atom);
        }
      }
    });
  } else if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
    console.warn('[Bug] could not find atom state to mount', atom);
  } // mount self


  var mounted = {
    d: new Set(initialDependent && [initialDependent]),
    l: new Set(),
    u: undefined
  };
  state.m.set(atom, mounted);
  state.u(function (prev) {
    if (isActuallyWritableAtom(atom) && atom.onMount) {
      var setAtom = function setAtom(update) {
        return writeAtom(state, atom, update);
      };

      mounted.u = atom.onMount(setAtom);
    }

    return prev;
  });
  return mounted;
};

var unmountAtom = function unmountAtom(state, wip, atom) {
  var _state$m$get;

  // unmount self
  var onUnmount = (_state$m$get = state.m.get(atom)) == null ? void 0 : _state$m$get.u;

  if (onUnmount) {
    onUnmount();
  }

  state.m.delete(atom); // unmount dependencies afterward

  var atomState = getAtomState(state, wip, atom);

  if (atomState) {
    if (atomState.p && typeof process === 'object' && process.env.NODE_ENV !== 'production') {
      console.warn('[Bug] deleting atomState with read promise', atom);
    }

    atomState.d.forEach(function (_, a) {
      if (a !== atom) {
        var mounted = state.m.get(a);

        if (mounted) {
          mounted.d.delete(atom);

          if (canUnmountAtom(a, mounted)) {
            unmountAtom(state, wip, a);
          }
        }
      }
    });
  } else if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
    console.warn('[Bug] could not find atom state to unmount', atom);
  }
};

var mountDependencies = function mountDependencies(state, wip) {
  wip.forEach(function (atomState, atom) {
    var _state$a$get;

    var prevDependencies = (_state$a$get = state.a.get(atom)) == null ? void 0 : _state$a$get.d;

    if (prevDependencies === atomState.d) {
      return;
    }

    var dependencies = new Set(atomState.d.keys());

    if (prevDependencies) {
      prevDependencies.forEach(function (_, a) {
        var mounted = state.m.get(a);

        if (dependencies.has(a)) {
          // not changed
          dependencies.delete(a);
        } else if (mounted) {
          mounted.d.delete(atom);

          if (canUnmountAtom(a, mounted)) {
            unmountAtom(state, wip, a);
          }
        } else if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
          console.warn('[Bug] a dependency is not mounted', a);
        }
      });
    }

    dependencies.forEach(function (a) {
      var mounted = state.m.get(a);

      if (mounted) {
        var dependents = mounted.d;
        dependents.add(atom);
      } else {
        mountAtom(state, wip, a, atom);
      }
    });
  });
};

var commitState = function commitState(state, wip) {
  // copy wip to AtomStateMap
  wip.forEach(function (atomState, atom) {
    if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
      Object.freeze(atomState);
    }

    var isNewAtom = state.n && !state.a.has(atom);
    state.a.set(atom, atomState);

    if (isNewAtom) {
      state.n(atom);
    }
  }); // empty wip

  wip.clear();
};

var subscribeAtom = function subscribeAtom(state, atom, callback) {
  var mounted = addAtom(state, atom);
  var listeners = mounted.l;
  listeners.add(callback);
  return function () {
    listeners.delete(callback);
    delAtom(state, atom);
  };
};

/*
export {
  unstable_createMutableSource as createMutableSource,
  unstable_useMutableSource as useMutableSource,
} from 'react'
*/
var TARGET = Symbol();
var GET_VERSION = Symbol();
var createMutableSource = function createMutableSource(target, getVersion) {
  var _ref;

  return _ref = {}, _ref[TARGET] = target, _ref[GET_VERSION] = getVersion, _ref;
};
var useMutableSource = function useMutableSource(source, getSnapshot, subscribe) {
  var lastVersion = react.useRef(0);
  var currentVersion = source[GET_VERSION](source[TARGET]);

  var _useState = react.useState(function () {
    return [
    /* [0] */
    source,
    /* [1] */
    getSnapshot,
    /* [2] */
    subscribe,
    /* [3] */
    currentVersion,
    /* [4] */
    getSnapshot(source[TARGET])];
  }),
      state = _useState[0],
      setState = _useState[1];

  var currentSnapshot = state[4];

  if (state[0] !== source || state[1] !== getSnapshot || state[2] !== subscribe || currentVersion !== state[3] && currentVersion !== lastVersion.current) {
    currentSnapshot = getSnapshot(source[TARGET]);
    setState([
    /* [0] */
    source,
    /* [1] */
    getSnapshot,
    /* [2] */
    subscribe,
    /* [3] */
    currentVersion,
    /* [4] */
    currentSnapshot]);
  }

  react.useEffect(function () {
    var didUnsubscribe = false;

    var checkForUpdates = function checkForUpdates() {
      if (didUnsubscribe) {
        return;
      }

      try {
        var nextSnapshot = getSnapshot(source[TARGET]);
        var nextVersion = source[GET_VERSION](source[TARGET]);
        lastVersion.current = nextVersion;
        setState(function (prev) {
          if (prev[0] !== source || prev[1] !== getSnapshot || prev[2] !== subscribe) {
            return prev;
          }

          if (prev[4] === nextSnapshot) {
            return prev;
          }

          return [
          /* [0] */
          prev[0],
          /* [1] */
          prev[1],
          /* [2] */
          prev[2],
          /* [3] */
          nextVersion,
          /* [4] */
          nextSnapshot];
        });
      } catch (e) {
        // schedule update
        setState(function (prev) {
          return [].concat(prev);
        });
      }
    };

    var unsubscribe = subscribe(source[TARGET], checkForUpdates);
    checkForUpdates();
    return function () {
      didUnsubscribe = true;
      unsubscribe();
    };
  }, [source, getSnapshot, subscribe]);
  return currentSnapshot;
};

var createStore = function createStore(initialValues, newAtomReceiver) {
  var state = createState(initialValues, newAtomReceiver);
  var mutableSource = createMutableSource(state, function () {
    return state.v;
  });

  var updateAtom = function updateAtom(atom, update) {
    return writeAtom(state, atom, update);
  };

  return [mutableSource, updateAtom];
}; // export only for tests/error.test.tsx, FIXME why do we need this?

var StoreContextMap = new Map();
var getStoreContext = function getStoreContext(scope) {
  if (!StoreContextMap.has(scope)) {
    StoreContextMap.set(scope, react.createContext(createStore()));
  }

  return StoreContextMap.get(scope);
};

var Provider = function Provider(_ref) {
  var initialValues = _ref.initialValues,
      scope = _ref.scope,
      children = _ref.children;
  var storeRef = react.useRef(null);

  if (typeof process === 'object' && process.env.NODE_ENV !== 'production' && process.env.NODE_ENV !== 'test') {
    /* eslint-disable react-hooks/rules-of-hooks */
    var atomsRef = react.useRef([]);

    if (storeRef.current === null) {
      // lazy initialization
      storeRef.current = createStore(initialValues, function (newAtom) {
        atomsRef.current.push(newAtom);
      });
    }

    useDebugState(storeRef.current, atomsRef.current);
    /* eslint-enable react-hooks/rules-of-hooks */
  } else {
    if (storeRef.current === null) {
      // lazy initialization
      storeRef.current = createStore(initialValues);
    }
  }

  var StoreContext = getStoreContext(scope);
  return react.createElement(StoreContext.Provider, {
    value: storeRef.current
  }, children);
};

var atomToPrintable = function atomToPrintable(atom) {
  return atom.debugLabel || atom.toString();
};

var stateToPrintable = function stateToPrintable(_ref2) {
  var state = _ref2[0],
      atoms = _ref2[1];
  return Object.fromEntries(atoms.flatMap(function (atom) {
    var mounted = state.m.get(atom);

    if (!mounted) {
      return [];
    }

    var dependents = mounted.d;
    var atomState = state.a.get(atom) || {};
    return [[atomToPrintable(atom), {
      value: atomState.e || atomState.p || atomState.w || atomState.v,
      dependents: Array.from(dependents).map(atomToPrintable)
    }]];
  }));
};

var getState = function getState(state) {
  return _extends({}, state);
}; // shallow copy
// We keep a reference to the atoms in Provider's atomsRef in dev mode,
// so atoms aren't garbage collected by the WeakMap of mounted atoms


var useDebugState = function useDebugState(store, atoms) {
  var subscribe = react.useCallback(function (state, callback) {
    // FIXME we don't need to resubscribe, just need to subscribe for new one
    var unsubs = atoms.map(function (atom) {
      return subscribeAtom(state, atom, callback);
    });
    return function () {
      unsubs.forEach(function (unsub) {
        return unsub();
      });
    };
  }, [atoms]);
  var state = useMutableSource(store[0], getState, subscribe);
  react.useDebugValue([state, atoms], stateToPrintable);
};

var keyCount = 0; // global key count for all atoms

function atom(read, write) {
  var key = "atom" + ++keyCount;
  var config = {
    toString: function toString() {
      return key;
    }
  };

  if (typeof read === 'function') {
    config.read = read;
  } else {
    config.init = read;

    config.read = function (get) {
      return get(config);
    };

    config.write = function (get, set, update) {
      set(config, typeof update === 'function' ? update(get(config)) : update);
    };
  }

  if (write) {
    config.write = write;
  }

  return config;
}

var isWritable = function isWritable(atom) {
  return !!atom.write;
};

function useAtom(atom) {
  var getAtomValue = react.useCallback(function (state) {
    var atomState = readAtom(state, atom);

    if (atomState.e) {
      throw atomState.e; // read error
    }

    if (atomState.p) {
      throw atomState.p; // read promise
    }

    if (atomState.w) {
      throw atomState.w; // write promise
    }

    if ('v' in atomState) {
      return atomState.v;
    }

    throw new Error('no atom value');
  }, [atom]);
  var subscribe = react.useCallback(function (state, callback) {
    return subscribeAtom(state, atom, callback);
  }, [atom]);
  var StoreContext = getStoreContext(atom.scope);

  var _useContext = react.useContext(StoreContext),
      mutableSource = _useContext[0],
      updateAtom = _useContext[1];

  var value = useMutableSource(mutableSource, getAtomValue, subscribe);
  var setAtom = react.useCallback(function (update) {
    if (isWritable(atom)) {
      return updateAtom(atom, update);
    } else {
      throw new Error('not writable atom');
    }
  }, [updateAtom, atom]);
  react.useDebugValue(value);
  return [value, setAtom];
}

exports.Provider = Provider;
exports.SECRET_INTERNAL_getStoreContext = getStoreContext;
exports.atom = atom;
exports.useAtom = useAtom;
